@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use base qw( IO::Socket );
 
-our $VERSION = '0.05';
+our $VERSION = '0.06';
 
 use Carp;
 
@@ -22,8 +22,10 @@ use Socket::GetAddrInfo qw(
 use Socket qw(
    SOCK_DGRAM
    SOL_SOCKET
-   SO_REUSEADDR SO_REUSEPORT SO_BROADCAST
+   SO_REUSEADDR SO_REUSEPORT SO_BROADCAST SO_ERROR
 );
+use POSIX qw( dup2 );
+use Errno qw( EINPROGRESS );
 
 my $IPv6_re = do {
    # translation of RFC 3986 3.2.2 ABNF to re
@@ -117,6 +119,14 @@ For symmetry with the accessor methods and compatibility with
 C<IO::Socket::INET>, C<PeerAddr> and C<PeerPort> are accepted as synonyms
 respectively.
 
+=item PeerAddrInfo => ARRAY
+
+Alternate form of specifying the peer to C<connect()> to. This should be an
+array of the form returned by C<Socket::GetAddrInfo::getaddrinfo>.
+
+This parameter overrides C<PeerHost> and C<PeerService>, and will ignore any
+values given for C<Type> or C<Proto>.
+
 =item Listen => INT
 
 Puts the socket into listening mode where new connections can be accepted
@@ -132,6 +142,14 @@ For symmetry with the accessor methods and compatibility with
 C<IO::Socket::INET>, C<LocalAddr> and C<LocalPort> are accepted as synonyms
 respectively.
 
+=item LocalAddrInfo => ARRAY
+
+Alternate form of specifying the local address to C<bind()> to. This should be
+an array of the form returned by C<Socket::GetAddrInfo::getaddrinfo>.
+
+This parameter overrides C<LocalHost> and C<LocalService>, and will ignore any
+values given for C<Type> or C<Proto>.
+
 =item ReuseAddr => BOOL
 
 If true, set the C<SO_REUSEADDR> sockopt
@@ -144,6 +162,11 @@ If true, set the C<SO_REUSEPORT> sockopt (not all OSes implement this sockopt)
 
 If true, set the C<SO_BROADCAST> sockopt
 
+=item Blocking => BOOL
+
+If false, the socket will be set to nonblocking mode. If absent or true it
+will be in blocking mode. See C<NON-BLOCKING> below for more detail.
+
 =back
 
 If the constructor fails, it will set C<$@> to an appropriate error message;
@@ -170,7 +193,7 @@ service name and number, in the form
 In this case, the name will be tried first, but if the resolver does not
 understand it then the port number will be used instead.
 
-=head1 $sock = IO::Socket::IP->new( $peeraddr )
+=head2 $sock = IO::Socket::IP->new( $peeraddr )
 
 As a special case, if the constructor is passed a single argument (as
 opposed to an even-sized list of key/value pairs), it is taken to be the value
@@ -243,7 +266,10 @@ sub configure
       $hints{protocol} = $proto;
    }
 
-   if( defined $arg->{LocalHost} or defined $arg->{LocalService} ) {
+   if( my $info = delete $arg->{LocalAddrInfo} ) {
+      @localinfos = @$info;
+   }
+   elsif( defined $arg->{LocalHost} or defined $arg->{LocalService} ) {
       # Either may be undef
       my $host = delete $arg->{LocalHost};
       my $service = delete $arg->{LocalService};
@@ -260,7 +286,10 @@ sub configure
       $err and ( $@ = "$err", return );
    }
 
-   if( defined $arg->{PeerHost} or defined $arg->{PeerService} ) {
+   if( my $info = delete $arg->{PeerAddrInfo} ) {
+      @peerinfos = @$info;
+   }
+   elsif( defined $arg->{PeerHost} or defined $arg->{PeerService} ) {
       defined( my $host = delete $arg->{PeerHost} ) or
          croak "Expected 'PeerHost'";
       defined( my $service = delete $arg->{PeerService} ) or
@@ -286,15 +315,14 @@ sub configure
 
    croak "Cannot Listen with a PeerHost" if defined $listenqueue and @peerinfos;
 
+   my $blocking = delete $arg->{Blocking};
+   defined $blocking or $blocking = 1;
+
    keys %$arg and croak "Unexpected keys - " . join( ", ", sort keys %$arg );
 
-   my $socketerr;
-   my $binderr;
-   my $connecterr;
-
+   my @infos;
    foreach my $local ( @localinfos ? @localinfos : {} ) {
       foreach my $peer ( @peerinfos ? @peerinfos : {} ) {
-
          next if defined $local->{family}   and defined $peer->{family}   and
             $local->{family} != $peer->{family};
          next if defined $local->{socktype} and defined $peer->{socktype} and
@@ -306,33 +334,106 @@ sub configure
          my $socktype = $local->{socktype} || $peer->{socktype} or next;
          my $protocol = $local->{protocol} || $peer->{protocol};
 
-         $self->socket( $family, $socktype, $protocol ) or ( $socketerr = $!, next );
-
-         foreach my $sockopt ( @sockopts_enabled ) {
-            $self->setsockopt( SOL_SOCKET, $sockopt, pack "i", 1 ) or ( $@ = "$!", return );
-         }
-
-         if( defined( my $addr = $local->{addr} ) ) {
-            $self->bind( $addr ) or ( $binderr = $!, next );
-         }
-
-         if( defined $listenqueue ) {
-            $self->listen( $listenqueue ) or ( $@ = "$!", return );
-         }
-
-         if( defined( my $addr = $peer->{addr} ) ) {
-            $self->connect( $addr ) or ( $connecterr = $!, next );
-         }
-
-         return $self;
+         push @infos, {
+            family    => $family,
+            socktype  => $socktype,
+            protocol  => $protocol,
+            localaddr => $local->{addr},
+            peeraddr  => $peer->{addr},
+         };
       }
    }
 
+   # In the nonblocking case, caller will be calling ->setup multiple times.
+   # Store configuration in the object for the ->setup method
+   # Yes, these are messy. Sorry, I can't help that...
+
+   ${*$self}{io_socket_ip_infos} = \@infos;
+
+   ${*$self}{io_socket_ip_idx} = -1;
+
+   ${*$self}{io_socket_ip_sockopts} = \@sockopts_enabled;
+   ${*$self}{io_socket_ip_listenqueue} = $listenqueue;
+   ${*$self}{io_socket_ip_blocking} = $blocking;
+
+   ${*$self}{io_socket_ip_errors} = [ undef, undef, undef ];
+
+   if( $blocking ) {
+      $self->setup or return undef;
+   }
+   return $self;
+}
+
+sub setup
+{
+   my $self = shift;
+
+   while(1) {
+      ${*$self}{io_socket_ip_idx}++;
+      last if ${*$self}{io_socket_ip_idx} >= @{ ${*$self}{io_socket_ip_infos} };
+
+      my $info = ${*$self}{io_socket_ip_infos}->[${*$self}{io_socket_ip_idx}];
+
+      $self->socket( @{$info}{qw( family socktype protocol )} ) or
+         ( ${*$self}{io_socket_ip_errors}[2] = $!, next );
+
+      $self->blocking( 0 ) unless ${*$self}{io_socket_ip_blocking};
+
+      foreach my $sockopt ( @{ ${*$self}{io_socket_ip_sockopts} } ) {
+         $self->setsockopt( SOL_SOCKET, $sockopt, pack "i", 1 ) or ( $@ = "$!", return undef );
+      }
+
+      if( defined( my $addr = $info->{localaddr} ) ) {
+         $self->bind( $addr ) or
+            ( ${*$self}{io_socket_ip_errors}[1] = $!, next );
+      }
+
+      if( defined( my $listenqueue = ${*$self}{io_socket_ip_listenqueue} ) ) {
+         $self->listen( $listenqueue ) or ( $@ = "$!", return undef );
+      }
+
+      if( defined( my $addr = $info->{peeraddr} ) ) {
+         # It seems that IO::Socket hides EINPROGRESS errors, making them look
+         # like a success. This is annoying here.
+         if( $self->connect( $addr ) ) {
+            $! = EINPROGRESS, return 0 if !$self->connected; # EINPROGRESS
+
+            $! = 0;
+            return 1;
+         }
+
+         return 0 if $! == EINPROGRESS;
+
+         ${*$self}{io_socket_ip_errors}[0] = $!;
+         next;
+      }
+
+      return 1;
+   }
+
+   $self->close;
+
    # Pick the most appropriate error, stringified
-   $@ = ( $connecterr || $binderr || $socketerr ) . '';
+   $! = ( grep defined, @{ ${*$self}{io_socket_ip_errors}} )[0];
+   $@ = "$!";
    return undef;
 }
 
+sub connect
+{
+   my $self = shift;
+   return $self->SUPER::connect( @_ ) if @_;
+
+   $! = 0, return 1 if $self->fileno and defined $self->peername;
+
+   if( $self->fileno ) {
+      # A connect has just failed, get its error value
+      ${*$self}{io_socket_ip_errors}[0] = $self->getsockopt( SOL_SOCKET, SO_ERROR );
+   }
+
+   return $self->setup;
+}
+
 =head1 METHODS
 
 =cut
@@ -453,11 +554,78 @@ sub accept
                     : $new;
 }
 
+# This second unbelievably dodgy hack guarantees that $self->fileno doesn't
+# change, which is useful during nonblocking connect
+sub socket
+{
+   my $self = shift;
+   return $self->SUPER::socket(@_) if not defined $self->fileno;
+
+   # I hate core prototypes sometimes...
+   CORE::socket( my $tmph, $_[0], $_[1], $_[2] ) or return undef;
+
+   dup2( $tmph->fileno, $self->fileno ) or die "Unable to dup2 $tmph onto $self - $!";
+}
+
 # Keep perl happy; keep Britain tidy
 1;
 
 __END__
 
+=head1 NON-BLOCKING
+
+If the constructor is passed a false value for the C<Blocking> argument, then
+the socket is put into nonblocking mode. When in nonblocking mode, an actual
+C<connect> operation will not have been completed by the time the constructor
+returns, because this may block.
+
+In order to use this mode, the caller should poll for writeability on the
+filehandle, each time it indicates write-readiness, the user code should call
+the C<connect> method, with no arguments. Once the socket has been connected
+to the peer, C<connect> will return true. If it returns false, the value of
+C<$!> indicates whether it should be tried again (C<EINPROGRESS>), or whether
+a permanent error has occured. This API is an extension of the
+C<IO::Socket::INET> API, unique to C<IO::Socket::IP>, because the former does
+not support multi-homed nonblocking connect.
+
+ use IO::Socket::IP;
+ use Errno qw( EINPROGRESS );
+ use Socket qw( SOCK_STREAM );
+
+ my $socket = IO::Socket::IP->new(
+    PeerHost    => "192.168.1.1",
+    PeerService => "25",
+    Type        => SOCK_STREAM,
+    Blocking    => 0,
+ ) or die "Cannot construct socket - $@";
+
+ while( !$socket->connect and $! == EINPROGRESS ) {
+    my $wvec = '';
+    vec( $wvec, fileno $socket, 1 ) = 1;
+
+    select( undef, $wvec, undef, undef ) or die "Cannot select - $!";
+ }
+
+ die "Cannot connect - $!" if $!;
+
+ ...
+
+This example uses C<select()>, but any similar mechanism should work
+analogously. C<IO::Socket::IP> takes care when creating new socket filehandles
+to preserve the actual file descriptor number, so such techniques as C<poll>
+or C<epoll> should be transparent to its reallocation of a different socket
+underneath, perhaps in order to switch protocol family between C<PF_INET> and
+C<PF_INET6>.
+
+When performing a non-blocking connect, it may be preferred to use the
+alternative C<LocalAddrInfo> and C<PeerAddrInfo> arguments. These allow the
+caller to perform the C<getaddrinfo> calls required, because if
+C<IO::Socket::IP> makes them they will definitely block. These arguments are
+provided for the case where the caller already has the C<getaddrinfo> results,
+perhaps obtained by some non-blocking method like L<Net::LibAsyncNS>, or by
+performing the lookup in a child process and passing the results back to the
+main process in an event loop of some kind.
+
 =head1 TODO
 
 =over 8
@@ -471,10 +639,31 @@ double-lookup overhead in such code as
 
 =item *
 
-Implement constructor args C<Timeout>, C<Blocking> and maybe C<Domain>. Except
-that C<Domain> is harder because L<IO::Socket> wants to dispatch to subclasses
+Implement constructor args C<Timeout> and maybe C<Domain>. Except that
+C<Domain> is harder because L<IO::Socket> wants to dispatch to subclasses
 based on it. Maybe C<Family> might be a better name?
 
+=item *
+
+Investigate whether C<POSIX::dup2> upsets BSD's C<kqueue> watchers, and if so,
+consider what possible workarounds might be applied.
+
+=back
+
+=head1 BUGS
+
+=over 8
+
+=item *
+
+Nonblocking connect fails unit tests on MSWin32 smoke-testing machines. The
+specifics of the failure are that C<connect()> seems to block anyway despite
+being asked not to, and that failure to connect is not detected properly. I am
+as yet unsure why this is.
+
+Blocking connect on MSWin32, and both blocking and nonblocking connect on
+other platforms, all test OK on smoke testing.
+
 =back
 
 =head1 AUTHOR
